import { SessionContext, sessionContextDialogs, DOMUtils, } from '@jupyterlab/apputils';
import { ActivityMonitor } from '@jupyterlab/coreutils';
import { pythonIcon } from '@jupyterlab/ui-components';
import { UUID } from '@lumino/coreutils';
import { Panel } from '@lumino/widgets';
import { flattenDeep, range } from 'lodash';
import { BytecodeModel } from './model';
import { escapeComments } from './utils';
import { BytecodeView } from './view';
/**
 * The panel that shows the Bytecode preview
 */
export class PythonBytecodePanel extends Panel {
    /**
     * Constructor for the panel.
     *
     * @param options: the instantiation options.
     */
    constructor(options) {
        var _a;
        super();
        this.id = DOMUtils.createDomID();
        this.title.label = 'Python Bytecode';
        this.title.closable = true;
        this.title.icon = pythonIcon;
        const { path, serviceManager, docManager, themeManager, userSettings, selections, } = options;
        const widget = docManager.findWidget(path);
        this._fileContext = docManager.contextForWidget(widget);
        this._docManager = docManager;
        this._themeManager = themeManager;
        this._selections = selections;
        const { kernelLanguagePreference, kernelAutoStart, startNewKernel, } = userSettings;
        const name = options.name || widget.context.contentsModel.name;
        this._sessionContext = new SessionContext({
            sessionManager: serviceManager.sessions,
            specsManager: serviceManager.kernelspecs,
            path: startNewKernel ? UUID.uuid4() : path,
            name: name || `Python Bytecode`,
            type: 'console',
            kernelPreference: {
                language: kernelLanguagePreference,
                autoStartDefault: (_a = kernelAutoStart) !== null && _a !== void 0 ? _a : true,
            },
        });
        this._monitor = new ActivityMonitor({
            signal: this._fileContext.model.contentChanged,
            timeout: 500,
        });
        this._model = new BytecodeModel();
        this._view = new BytecodeView(this._model);
        this.addWidget(this._view);
    }
    /**
     * Get the model.
     */
    get model() {
        return this._model;
    }
    /**
     * Get the session context.
     */
    get sessionContext() {
        return this._sessionContext;
    }
    /**
     * Dispose the panel.
     */
    dispose() {
        // TODO: dispose session if last panel disposed?
        this._removeListeners();
        this._sessionContext.dispose();
        this._view.dispose();
        super.dispose();
    }
    /**
     * Setup the panel.
     */
    async setup() {
        var _a;
        const value = await this._sessionContext.initialize();
        if (value) {
            await sessionContextDialogs.selectKernel(this._sessionContext);
        }
        const { name, kernelDisplayName } = this._sessionContext;
        if (!name.endsWith(kernelDisplayName)) {
            await ((_a = this._sessionContext.session) === null || _a === void 0 ? void 0 : _a.setName(`${name} - ${kernelDisplayName}`));
        }
        this.title.label = `${this._sessionContext.name} Bytecode`;
        this.title.caption = this._sessionContext.name;
        this._setupListeners();
        // do not block on first request
        this._getFileContent();
        this._changeTheme();
    }
    /**
     * Handle onCloseRequest messages.
     * @param msg The message
     */
    onCloseRequest(msg) {
        super.onCloseRequest(msg);
        this.dispose();
    }
    /**
     * Setup the listeners.
     */
    _setupListeners() {
        this._monitor.activityStopped.connect(this._getModelContent, this);
        this._fileContext.fileChanged.connect(this._getFileContent, this);
        this._fileContext.disposed.connect(this.dispose, this);
        this._selections.changed.connect(this._handleSelectionChanged, this);
        this._sessionContext.kernelChanged.connect(this._handleKernelChanged, this);
        // TODO: make themeManager optional
        if (this._themeManager) {
            this._themeManager.themeChanged.connect(this._changeTheme, this);
        }
    }
    /**
     * Remove the listeners.
     */
    _removeListeners() {
        if (this._monitor) {
            this._monitor.activityStopped.disconnect(this._getModelContent, this);
            this._monitor.dispose();
        }
        if (this._fileContext) {
            this._fileContext.fileChanged.disconnect(this._getFileContent, this);
            this._fileContext.disposed.disconnect(this.dispose, this);
        }
        if (this._themeManager) {
            this._themeManager.themeChanged.disconnect(this._changeTheme, this);
        }
        if (this._selections) {
            this._selections.changed.disconnect(this._handleSelectionChanged, this);
        }
        this._sessionContext.kernelChanged.disconnect(this._handleKernelChanged, this);
    }
    /**
     * Evaluate the content of the document.
     * @param content the content to evaluate
     */
    _evaluateContent(content) {
        const msg = Private.formatKernelMessage(content);
        return this._execute(msg);
    }
    /**
     * Get the content of the model.
     */
    async _getModelContent() {
        const content = this._fileContext.model.toString();
        return this._evaluateContent(content);
    }
    /**
     * Get the content of the file.
     */
    async _getFileContent() {
        const path = this._fileContext.path;
        const file = await this._docManager.services.contents.get(path);
        return this._evaluateContent(file.content);
    }
    /**
     * Execute code.
     * @param code The code to execute
     */
    async _execute(code) {
        var _a, _b;
        const future = (_b = (_a = this._sessionContext.session) === null || _a === void 0 ? void 0 : _a.kernel) === null || _b === void 0 ? void 0 : _b.requestExecute({
            code,
        });
        future.onIOPub = this._model.handleKernelMessage;
        return future.done;
    }
    /**
     * Change the theme to the current theme.
     */
    _changeTheme() {
        if (!this._themeManager) {
            return;
        }
        const isLight = this._themeManager.isLight(this._themeManager.theme);
        this._model.isLight = isLight;
    }
    /**
     * Handle a kernel changed event.
     */
    _handleKernelChanged() {
        var _a;
        if (!((_a = this._sessionContext.session) === null || _a === void 0 ? void 0 : _a.kernel)) {
            this.dispose();
            return;
        }
    }
    /**
     * Handle a seleection changed event.
     */
    _handleSelectionChanged() {
        const selectedLines = flattenDeep(this._selections.values().map(s => s.map(e => {
            const [start, end] = [e.start, e.end].sort((a, b) => a.line - b.line);
            let [startLine, endLine] = [start.line, end.line];
            if (startLine != endLine && end.column === 0) {
                endLine--;
            }
            return range(startLine, endLine + 1);
        })));
        this.model.selectedLines = new Set(selectedLines);
    }
}
var Private;
(function (Private) {
    /**
     * Format the kernel message.
     * @param code the content
     */
    function formatKernelMessage(code) {
        const escapedContent = escapeComments(code);
        return ['import dis', `dis.dis("""${escapedContent}""")`].join('\n');
    }
    Private.formatKernelMessage = formatKernelMessage;
})(Private || (Private = {}));
